#pragma once



#include "problem/transformation.hpp"
#include "problem/problem.hpp"
#include "problem/bbob.hpp"
#include "problem/pbo.hpp"

// #include "problem/python.hpp"
#include "problem/wmodel.hpp"
